-- (c) David Cunningham 2009, Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php

print("Loading wheel_assembly.lua")

-- The wheel assembly handles physical simulation and rendering of a wheel + suspension

WheelAssembly = WheelAssembly or { }

function WheelAssembly.new (obj, aim_bone, steer_bone, axle_bone, wheel_bone,
                            wheel_radius, slack, mass, mu, off_road,
                            max_restitution, damping_factor)
        local self = {}
        make_instance(self,WheelAssembly)

        -- these guys set and maintained later
        self.onGround = false
        self.skidding = 0
        self.locked = false
        self.tractionControl = 1
        self.travel = 0 -- gets converted to rotation
        self.steer = 0
        self.freeWheelSpeed = 0
        self.rayPosOS = Vector3(0,0,0)
        self.rayDirOS = Vector3(0,0,0)
        self.axleDirBeforeSteerOS = Vector3(0,0,0)
        self.axleDirOS = Vector3(0,0,0)

        self.restitutionForce = Vector3(0,0,0)
        self.velocityDifference = Vector3(0,0,0)
        self.worldContactPos = Vector3(0,0,0)
        self.worldContactNormal = Vector3(0,0,0)

        -- graphics
        self.obj = obj
        self.parentEntity = obj.entity
        self.aimBone = self.parentEntity:getBoneId(aim_bone or steer_bone)
        if steer_bone then
                self.steerBone = self.parentEntity:getBoneId(steer_bone)
        end
        self.axleBone = self.parentEntity:getBoneId(axle_bone)
        self.wheelBone = self.parentEntity:getBoneId(wheel_bone)

        self.wheelRadius = wheel_radius
        self.wheelCircumference = 2 * wheel_radius * math.pi
        self.mu = mu
        self.offRoad = off_road or false
        self.maxRestitution = max_restitution --perfectly ok for this to be nil

        self:updatePos()

        local hooke = 2 * (9.81*mass) / self.suspensionLength
        local damping = 2 * math.sqrt(mass * hooke) * (damping_factor or 0.5)
        self.suspensionHooke = hooke
        self.suspensionDamping = damping
         -- probably when vehicles are spawned, they are 'parked' so this is a reasonable initial 'previous' value for computing damping
        self.lastExtension = -self.suspensionLength/2

        self:setTorque(0)

        return self
end

function WheelAssembly:destroy ()
end

-- update based on changes to the wheel attachment to the parent vehicle, e.g. retracting gear on planes
-- use the unmodified bone transforms to figure out some of the needed metrics
function WheelAssembly:updatePos(amount)
        local px,py,pz, qw,qx,qy,qz = self.parentEntity:getBoneDerivedPositionOrientation(self.aimBone)
        self.rayPosOS:xyz(px,py,pz)
        self.rayDirOS:xyz(mulqv(qw,qx,qy,qz, 0,1,0)) -- ray is +y in bone space
        --qw,qx,qy,qz = self.parentEntity:getBoneInitialOrientation(self.steerBone) -- don't want the 'steer' effect yet
        local ww,wx,wy,wz = self.parentEntity:getBoneInitialOrientation(self.wheelBone)
        self.axleDirBeforeSteerOS:xyz(mulqv(qw,qx,qy,qz, mulqv(ww,wx,wy,wz, 0,1,0)))
        
        local _, ay, _ = self.parentEntity:getBoneInitialPosition(self.axleBone)
        local _, wy, _ = self.parentEntity:getBoneInitialPosition(self.wheelBone)
        self.slack = ay
        self.suspensionLength = wy

        self:setSteer(self.steer)
        
end

function WheelAssembly:setSteer(amount)
        self.steer = amount
        local dx,dy,dz = self.rayDirOS:xyz()
        local sw,sx,sy,sz = quat_angle(self.steer, dx,dy,dz)
        self.axleDirOS:xyz(mulqv(sw,sx,sy,sz, self.axleDirBeforeSteerOS:xyz()))
end

function WheelAssembly:setTorque(torque)
        self.driveForce = self.wheelRadius==0 and 0 or (torque / self.wheelRadius)
end

-- called from update callback, NOT step callback
function WheelAssembly:updateGFX()
        local e = self.parentEntity
        e:setBonePositionOffset(self.axleBone, 0, self.lastExtension, 0)
        if self.steerBone then
                e:setBoneOrientationOffsetAngle(self.steerBone, self.steer, 0,1,0)
        end
        local angle = 360 * self.travel / self.wheelCircumference
        e:setBoneOrientationOffsetAngle(self.wheelBone, angle, 0,-1,0)
end

function WheelAssembly:extensionChange(extension)
        local extension_delta = extension - self.lastExtension
        self.lastExtension = extension
        return extension_delta
end

function WheelAssembly:freeWheel (interval)
        if math.abs(self.driveForce) > 10 and not self.locked and self.obj.engine.gear ~= 0 then
                self.freeWheelSpeed = 15 * self.driveForce/math.abs(self.driveForce)
        else
                -- wind down
                self.freeWheelSpeed = self.freeWheelSpeed*math.pow(0.7,interval)
        end
        self.travel = (self.travel + self.freeWheelSpeed * interval) % self.wheelCircumference
end

function WheelAssembly:groundWheel (forward_speed, interval)
        local distance = forward_speed * interval
        self.travel = (self.travel + distance) % self.wheelCircumference
        self.freeWheelSpeed = distance / interval
end

function WheelAssembly:process (interval)

        local vehicle = self.obj
        local body = vehicle.body
        local world = vehicle.world

        -- _ws means world space

        -- ALLOCATION ALLOCATION
        local ray_reach, ground, nx, ny, nz, mat =
                body:cast(-1, self.rayPosOS,
                          self.rayDirOS,
                          self.suspensionLength + self.wheelRadius + self.slack,
                          true, false, body)

        self.onGround = ray_reach ~= nil

        ray_reach = ray_reach or self.suspensionLength + self.wheelRadius + self.slack

        -- contact point (world space)
        local cpx, cpy, cpz = self.rayDirOS:xyz()
        cpx, cpy, cpz = mulnv(ray_reach, cpx,cpy,cpz)
        cpx, cpy, cpz = addvv(cpx,cpy,cpz, self.rayPosOS:xyz())
        cpx, cpy, cpz = body:localToWorld(cpx,cpy,cpz)

        local extension = ray_reach - self.wheelRadius - self.suspensionLength - self.slack

        extension = clamp(extension, -self.suspensionLength, 0)


        local extension_rate = self:extensionChange(extension) / interval

        self.skidding = 0

        if not self.onGround then
                self:freeWheel(interval)
                return
        end


        local spring_force_mag   = - extension * self.suspensionHooke

        local damping_force_mag  = - extension_rate * self.suspensionDamping


        -- this is the important one
        local surface_force_mag = spring_force_mag + damping_force_mag
        local surface_force_mag_clamped = surface_force_mag

        if self.maxRestitution and surface_force_mag > self.maxRestitution then
                surface_force_mag_clamped = self.maxRestitution
        end


        local rx, ry, rz = mulnv(surface_force_mag_clamped, nx, ny, nz)

        --if player_ctrl.vehicle ~= nil and player_ctrl.vehicle.instance == self.obj then
        --        echo("RF: ",extension)
        --end
        body:force(rx,ry,rz, cpx,cpy,cpz)

        -- subtract ground velocity from body velocity
        local vdx, vdy, vdz = ground:getLocalVelocity(cpx, cpy, cpz, true)
        vdx, vdy, vdz = addvv(-vdx,-vdy,-vdz, body:getLocalVelocity(cpx, cpy, cpz, true))

        local trx, try, trz -- traction force
        local frx, fry, frz -- friction force

        if self.locked then

                -- ignore all but the 'along the ground' part
                vdx, vdy, vdz = project(vdx,vdy,vdz, nx,ny,nz)

                local speed_difference = lenv(vdx,vdy,vdz)

                local friction, drive = global_friction_model(vehicle, self.mu, self.offRoad, surface_force_mag,
                                                              speed_difference, 0, 1, mat)

                trx, try, trz = 0,0,0

                self.skidding = clamp((speed_difference-5)/35,0,1)

                if friction==0 then speed_difference = 1 end
                frx, fry, frz = mulnv(-friction/speed_difference, vdx,vdy,vdz) -- normalise and scale by friction

        else

                -- forward (object space)
                local fox,foy,foz = cross(nx,ny,nz, self.axleDirOS:xyz())

                -- right (object space)
                local rox,roy,roz = cross(nx,ny,nz, fox,foy,foz)

                -- transform to world space
                local fwx,fwy,fwz = body:rotateToWorld(fox,foy,foz)
                local rwx,rwy,rwz = body:rotateToWorld(rox,roy,roz)

                -- divide into ahead and sideways component
                local forward_speed = dot(vdx,vdy,vdz, fwx,fwy,fwz)
                local right_speed = dot(vdx,vdy,vdz, rwx,rwy,rwz)


                local friction, drive, spin = global_friction_model(vehicle, self.mu, self.offRoad, surface_force_mag,
                                                                    math.abs(right_speed), self.driveForce, self.tractionControl, mat)

                self.skidding = spin and 1 or 0

                if self.skidding < 1 then
                        self:groundWheel(forward_speed, interval)
                else
                        self:freeWheel(interval)
                end

                trx,try,trz = mulnv(drive, fwx,fwy,fwz)
                frx,fry,frz = mulnv(-sign(right_speed) * friction, rwx,rwy,rwz) -- scale by friction and correct direction
        end     

        body:force(frx,fry,frz, cpx,cpy,cpz)
        body:force(trx,try,trz, cpx,cpy,cpz)

        -- force on the ground
        local grx,gry,grz = addvv(rx,ry,rz, addvv(frx,fry,frz, trx,try,trz))

        -- hack to avoid applying extremely large forces
        local magnitude = math.min(lenv(grx,gry,grz), ground.mass * 10)
        grx, gry, grz = mulnv(-magnitude, normv(grx,gry,grz))
        ground:force(grx,gry,grz, cpx,cpy,cpz)

        self.worldContactPos:xyz(cpx,cpy,cpz)
        self.worldContactNormal:xyz(nx, ny, nz)
        self.velocityDifference:xyz(vdx,vdy,vdz)
        self.restitutionForce:xyz(rx,ry,rz)


end

-- vim: shiftwidth=8:tabstop=8:expandtab
